#include <cmath>

class Matriz {

    public:
    int numLinhas, numColunas;
    float** matriz;

    Matriz(int l, int c){
		
        this->numLinhas = l;
        this->numColunas = c;
		this->matriz = new float*[numLinhas];
		for(int i = 0; i < this->numLinhas; i++){
			this->matriz[i] = new float[numColunas]; 
            for(int j = 0; j < this->numColunas; j++){
                this->matriz[i][j] = 0.0f;
            }
        }
    }

    void setarElemento(int posL, int posC, float val){
        this->matriz[posL][posC] = val;
    }

    float getElemento(int l, int c){
        return this->matriz[l][c];
    }
};

class Ponto {

    public:

    float x, y, z;

    Ponto(float x, float y, float z){

        this->x = x;
        this->y = y;
        this->z = z;
    }
};


class Vetor {


    public:
    Ponto* inicio, *fim;
    float coord[4];

    Vetor(Ponto* i, Ponto* f){

        this->inicio = i;
        this->fim = f;

        this->coord[0] = this->fim->x - this->inicio->x;
        this->coord[1] = this->fim->y - this->inicio->y;
        this->coord[2] = this->fim->z - this->inicio->z;
        this->coord[3] = 1.0;

    }

    float getNorma(){
        return sqrt((this->coord[0] * this->coord[0]) + (this->coord[1] * this->coord[1]) + (this->coord[2] * this->coord[2]));
    }

};

class Reta {

    public:
    Vetor* vetor;
    Ponto* inicio;
	
    Reta(Vetor* v, Ponto* p){
        this->vetor = v;
        this->inicio = p;
    }
	
	Ponto* getPonto(float t){
		Ponto* retorno = new Ponto(inicio->x + vetor->coord[0]*t,inicio->y + vetor->coord[1]*t,inicio->z + vetor->coord[2]*t);
		return retorno;
	}

	float getX(float t){
		return inicio->x + vetor->coord[0]*t;
	}

	float getY(float t){
		return inicio->y + vetor->coord[1]*t;
	}

	float getZ(float t){
		return inicio->z + vetor->coord[2]*t;
	}
};



class Triangulo {

    public:
    int indexA, indexB, indexC;

    Triangulo(int a, int b, int c){

        this->indexA = a;
        this->indexB = b;
        this->indexC = c;
    }

};

Ponto* projecaoPlanoVisao(float d, Ponto* ponto){
		float novoX = ponto->x*(d/ponto->z);
		float novoY = ponto->y*(d/ponto->z);
		Ponto* retorno = new Ponto(novoX,novoY,d);
		return retorno;
	}

	float produtoEscalar(Vetor* v1, Vetor* v2){
		return (v1->coord[0]*v2->coord[0])+v1->coord[1]*v2->coord[1] + v1->coord[2] * v2->coord[2];
	}
	
	float produtoEscalar(Ponto* p1, Vetor* v2){
		return (p1->x*v2->coord[0])+p1->y*v2->coord[1] + p1->z * v2->coord[2];
	}

	void produtoEscalar(float k, Vetor* v1){
		v1->coord[0] = v1->coord[0]*k;	
		v1->coord[1] = v1->coord[1]*k;
		v1->coord[2] = v1->coord[2]*k;
	}  

class Operacoes {

    public:

    Operacoes(){

    }

    Matriz* mult(Matriz* matA, Matriz* matB){

        Matriz* resultado = new Matriz(matA->numLinhas, matB->numColunas);

        if(matA->numColunas == matB->numLinhas){

            int a,b,c;

            for(a = 0; a < matA->numLinhas; a++){

                for(b = 0; b < matB->numColunas; b++){

                    float soma = 0.0;

                    for(c = 0; c < matB->numLinhas; c++){

                        soma = soma + (matA->matriz[a][c] * matB->matriz[c][b]);

                    }

                    resultado->setarElemento(a,b,soma);
                }
            }
        }

        return resultado;
    }

    Vetor* multPorVetor(Matriz* mat, Vetor* vet){
		
        int a,b;

        for(a = 0; a < mat->numLinhas; a++){

            float soma = 0.0;

            for(b = 0; b < 3; b++){

                soma += (mat->matriz[a][b] * vet->coord[b]);
            }

            vet->coord[a] = soma;
        }
        return vet;
    }

    Vetor* translacao(Vetor* vetor, float x, float y, float z){

        Matriz* matTrans = new Matriz(4,4);

        matTrans->setarElemento(0, 3, x);
        matTrans->setarElemento(1, 3, y);
        matTrans->setarElemento(2, 3, z);

        matTrans->setarElemento(0, 0, 1.0);
        matTrans->setarElemento(1, 1, 1.0);
        matTrans->setarElemento(2, 2, 1.0);
        matTrans->setarElemento(3, 3, 1.0);
/*
        Matriz* matVetor = new Matriz(4, 1);

        matVetor->setarElemento(0, 0, vetor->coord[0]);
        matVetor->setarElemento(1, 0, vetor->coord[1]);
        matVetor->setarElemento(2, 0, vetor->coord[2]);
        matVetor->setarElemento(3, 0, 1);

        matVetor = this->mult(matTrans, matVetor);

        vetor->x = matVetor->getElemento(0, 0);
        vetor->y = matVetor->getElemento(1, 0);
        vetor->z = matVetor->getElemento(2, 0);
*/
        vetor = multPorVetor(matTrans, vetor);
        return vetor;
    }

    Vetor* rotacaoX(Vetor* vetor, float angEmRad, bool positiva){

        Matriz* matRotacao = new Matriz(4, 4);

        matRotacao->setarElemento(0, 0, 1.0);
        matRotacao->setarElemento(3, 3, 1.0);

        if(positiva){

            matRotacao->setarElemento(1, 1, cos(angEmRad));
            matRotacao->setarElemento(1, 2, (-1 * sin(angEmRad)));
            matRotacao->setarElemento(2, 1, sin(angEmRad));
            matRotacao->setarElemento(2, 2, cos(angEmRad));

        }else {

            matRotacao->setarElemento(1, 1, cos(angEmRad));
            matRotacao->setarElemento(1, 2, sin(angEmRad));
            matRotacao->setarElemento(2, 1, (-1 * sin(angEmRad)));
            matRotacao->setarElemento(2, 2, cos(angEmRad));

        }
/*
        Matriz* matVetor = new Matriz(4, 1);

        matVetor->setarElemento(0, 0, vetor->x);
        matVetor->setarElemento(1, 0, vetor->y);
        matVetor->setarElemento(2, 0, vetor->z);
        matVetor->setarElemento(3, 0, 1);
*/
        vetor = this->multPorVetor(matRotacao, vetor);
/*
        vetor->x = matVetor->getElemento(0, 0);
        vetor->y = matVetor->getElemento(1, 0);
        vetor->z = matVetor->getElemento(2, 0);

*/
        return vetor;
    }

    Vetor* rotacaoY(Vetor* vetor, float angEmRad, bool positiva){

        Matriz* matRotacao = new Matriz(4, 4);

        matRotacao->setarElemento(1, 1, 1.0);
        matRotacao->setarElemento(3, 3, 1.0);

        if(positiva){

            matRotacao->setarElemento(0, 0, cos(angEmRad));
            matRotacao->setarElemento(0, 2, sin(angEmRad));
            matRotacao->setarElemento(2, 0, (-1 * sin(angEmRad)));
            matRotacao->setarElemento(2, 2, cos(angEmRad));

        }else {

            matRotacao->setarElemento(0, 0, cos(angEmRad));
            matRotacao->setarElemento(0, 2, (-1 *sin(angEmRad)));
            matRotacao->setarElemento(2, 0, sin(angEmRad));
            matRotacao->setarElemento(2, 2, cos(angEmRad));
        }

        vetor = this->multPorVetor(matRotacao, vetor);

        return vetor;
    }

    Vetor* rotacaoZ(Vetor* vetor, float angEmRad, bool positiva){
        Matriz* matRotacao = new Matriz(4, 4);

        matRotacao->setarElemento(2, 2, 1.0);
        matRotacao->setarElemento(3, 3, 1.0);

        if(positiva){

            matRotacao->setarElemento(0, 0, cos(angEmRad));
            matRotacao->setarElemento(0, 1, (-1 *sin(angEmRad)));
            matRotacao->setarElemento(1, 0, sin(angEmRad));
            matRotacao->setarElemento(1, 1, cos(angEmRad));

        }else {

            matRotacao->setarElemento(0, 0, cos(angEmRad));
            matRotacao->setarElemento(0, 1, sin(angEmRad));
            matRotacao->setarElemento(1, 0, (-1 * sin(angEmRad)));
            matRotacao->setarElemento(1, 1, cos(angEmRad));
        }

        vetor = this->multPorVetor(matRotacao, vetor);

        return vetor;
    }

    Vetor* mudancaDeEscala(Vetor* vetor, float novoX, float novoY, float novoZ){

        Matriz* matEscala = new Matriz(4, 4);

        matEscala->setarElemento(0, 0, novoX);
        matEscala->setarElemento(1, 1, novoY);
        matEscala->setarElemento(2, 2, novoZ);
        matEscala->setarElemento(3, 3, 1.0);

        vetor = multPorVetor(matEscala, vetor);

        return vetor;
    }

    Vetor* cisalhamentoXY(Vetor* vetor, float valor){

        Matriz* matCis = new Matriz(4, 4);

        matCis->setarElemento(0, 0, 1);
        matCis->setarElemento(1, 1, 1);
        matCis->setarElemento(2, 2, 1);
        matCis->setarElemento(3, 3, 1);

        matCis->setarElemento(0, 1, valor);

        vetor = multPorVetor(matCis, vetor);

        return vetor;
    }

    Vetor* cisalhamentoXZ(Vetor* vetor, float valor){

        Matriz* matCis = new Matriz(4, 4);

        matCis->setarElemento(0, 0, 1);
        matCis->setarElemento(1, 1, 1);
        matCis->setarElemento(2, 2, 1);
        matCis->setarElemento(3, 3, 1);

        matCis->setarElemento(0, 2, valor);

        vetor = multPorVetor(matCis, vetor);

        return vetor;
    }

    Vetor* cisalhamentoYZ(Vetor* vetor, float valor){

        Matriz* matCis = new Matriz(4, 4);

        matCis->setarElemento(0, 0, 1);
        matCis->setarElemento(1, 1, 1);
        matCis->setarElemento(2, 2, 1);
        matCis->setarElemento(3, 3, 1);

        matCis->setarElemento(1, 2, valor);

        vetor = multPorVetor(matCis, vetor);

        return vetor;

    }

    Vetor* reflexaoX(Vetor *vetor){

        Matriz* matReflexao = new Matriz(4, 4);

        matReflexao->setarElemento(0, 0, -1);
        matReflexao->setarElemento(1, 1, 1);
        matReflexao->setarElemento(2, 2, 1);
        matReflexao->setarElemento(3, 3, 1);

        vetor = multPorVetor(matReflexao, vetor);

        return vetor;
    }

    Vetor* reflexaoY(Vetor *vetor){

        Matriz* matReflexao = new Matriz(4, 4);

        matReflexao->setarElemento(0, 0, 1);
        matReflexao->setarElemento(1, 1, -1);
        matReflexao->setarElemento(2, 2, 1);
        matReflexao->setarElemento(3, 3, 1);

        vetor = multPorVetor(matReflexao, vetor);

        return vetor;

    }

    Vetor* reflexaoZ(Vetor *vetor){

        Matriz* matReflexao = new Matriz(4, 4);

        matReflexao->setarElemento(0, 0, 1);
        matReflexao->setarElemento(1, 1, 1);
        matReflexao->setarElemento(2, 2, -1);
        matReflexao->setarElemento(3, 3, 1);

        vetor = multPorVetor(matReflexao, vetor);

        return vetor;

    }

	
	Vetor* mudancaDeBase(Vetor* vetorBase,Vetor* u, Vetor* v, Vetor* n){
		//vetor x canonico
		Ponto* pX = new Ponto(1,0,0);
		Ponto* pXo = new Ponto (0,0,0);
		Vetor* x = new Vetor(pXo, pX);
		//vetor y canonico
		Ponto* pY = new Ponto(0,1,0);
		Ponto* pYo = new Ponto(0,0,0);
		Vetor* y = new Vetor(pYo, pY);
		//vetor Z canonico
		Ponto* pZ = new Ponto(0,0,1);
		Ponto* pZo = new Ponto(0,0,0);
		Vetor* z = new Vetor(pZo, pZ);
		//criando matriz de mudana de base
		Matriz* matrizMudancaBase = new Matriz(3,3);

		matrizMudancaBase->setarElemento(0,0,produtoEscalar(u,x));
		matrizMudancaBase->setarElemento(0,1,produtoEscalar(u,y));
		matrizMudancaBase->setarElemento(0,2,produtoEscalar(u,z));

		matrizMudancaBase->setarElemento(1,0,produtoEscalar(v,x));
		matrizMudancaBase->setarElemento(1,1,produtoEscalar(v,y));
		matrizMudancaBase->setarElemento(1,2,produtoEscalar(v,z));

		matrizMudancaBase->setarElemento(2,0,produtoEscalar(n,x));
		matrizMudancaBase->setarElemento(2,1,produtoEscalar(n,y));
		matrizMudancaBase->setarElemento(2,2,produtoEscalar(n,z));
		
		//criar a matriz do Ponto com base canonica
		Matriz* vetorMatriz = new Matriz(3,1);
		vetorMatriz->setarElemento(0,0,vetorBase->coord[0]);
		vetorMatriz->setarElemento(1,0,vetorBase->coord[1]);
		vetorMatriz->setarElemento(2,0,vetorBase->coord[2]);

		//matriz retorno
		Matriz* retorno = mult(matrizMudancaBase, vetorMatriz);
		
		//pegando as coordenandas
		float fimX = retorno->matriz[0][0]*u->coord[0] + retorno->matriz[1][0]*v->coord[0] + retorno->matriz[2][0]*n->coord[0];
		float fimY = retorno->matriz[0][0]*u->coord[1] + retorno->matriz[1][0]*v->coord[1] + retorno->matriz[2][0]*n->coord[1];
		float fimZ = retorno->matriz[0][0]*u->coord[2] + retorno->matriz[1][0]*v->coord[2] + retorno->matriz[2][0]*n->coord[2];
		
		//criar os pontos
		Ponto* f = new Ponto(fimX,fimY,fimZ);
		Ponto* ini = new Ponto(0,0,0);
		//vetor retorno
		Vetor* chegandoAoFinalDoMetodo = new Vetor(ini,f);
		return chegandoAoFinalDoMetodo;
	}

};

	